unit GWMonOLE;
{---------------------------------------------------------------
Purpose: Provides monitoring functions for observing behavior of OLE objects.

Author:  Graham Wideman
         graham@wideman-one.com
         www.wideman-one.com

How to incorporate into a project:
----------------------------------
1) Add this unit to project.  By default this will add code to Project Source
   to Auto-Create this form.  Disable this by deleting the Auto-Create line
   in Project Source, or equivalently removing this form from the Auto-Create
   list in Project Options dialog.

2) Create this form from application code -- typically another Form's
   CreateForm routine, giving that Form as the  Owner parameter. (This
   causes the owning form to automatically destroy our monitor form and
   data when the owner form is destroyed.)

3) When creating OLE objects that you want to monitor, inherit from
   TGWMonInterfacedObject, and provide the fm_OLEMonitor form (created in step 2)
   for the  Afm_OLEMonitor parameter in TGWMonInterfacedObject.Create.

Revision History
----------------
98-07-24 GW V1.1 first version as separate unit

To-do list
----------

---------------------------------------------------------------}

interface
uses Windows, ActiveX, Forms, grids, Classes, Controls, ExtCtrls, StdCtrls;

Const
  GW_MonCount      = 10;
  GW_MonDataCount  = 20;

Type
  TGW_Mon    = class;  // forward

  //----------------------------
  Tfm_OLEMonitor = class(TForm)
  //----------------------------
    sg_OLEMonitor: TStringGrid;
    Panel1: TPanel;
    tm_Update: TTimer;
    StaticText1: TStaticText;
    sh_update: TShape;
    bt_Clear: TButton;
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormCreate(Sender: TObject);
    procedure tm_UpdateTimer(Sender: TObject);
    procedure bt_ClearClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);

  private
    { Private declarations }
    FMons          : array[0..GW_MonCount-1] of TGW_Mon;

  protected
    // Messages from FMon's
    function    Mon_Add(AMon: TGW_Mon): integer; { returns monitor "handle" }
    procedure   Mon_Deleting(MonId: integer);

  public
    { Public declarations }
    procedure DockMe;

    // Messages from application
    procedure   ClearData; virtual;
    procedure   ClearDeadMonitors;
    procedure   GridUpdate;

    // Special test procedures
    procedure DeleteCOMObject(id: integer);
    procedure DeleteAllCOMObjects;
  end;

  //---------------------------------------
  TGWInterfacedObject = class(TObject, IUnknown)
  //---------------------------------------
  // Similar to interfaced object
  private
    FRefCount: Integer;
  protected
    function QueryInterface(const IID: TGUID; out Obj): Integer; stdcall;
    function _AddRef: Integer; stdcall;
    function _Release: Integer; stdcall;
  public
    property  RefCount: Integer read FRefCount;
    procedure RefCountChanged; virtual;
  end;

  //---------------------------------------
  TGWMonInterfacedObject = class(TGWInterfacedObject)
  //---------------------------------------
  protected
    FMyObjectId  : integer;
    FMon         : TGW_Mon;
  public
    constructor Create(Afm_OLEMonitor: Tfm_OLEMonitor); // or nil for no monitoring
    destructor  Destroy; override;
    procedure   RefCountChanged; override;

    procedure   Mon_Deleting;
    procedure   Mon_IncData(ItemNum: integer);
  end;


  //---------------------------------------
  TGW_Mon       = class
  //---------------------------------------
  protected
    Ffm_OLEMonitor : Tfm_OLEMonitor;            // who we report to
    FMonitoree     : TGWMonInterfacedObject;    // so we can get at RefCount_Exposed
    FMonId         : integer;                   // index in FMons array
    FData          : array[0..GW_MonDataCount-1] of integer;
    FRowName       : array[0..GW_MonDataCount-1] of string;
    FDataCount     : integer;
  public
    constructor Create(AMonitoree: TGWMonInterfacedObject; Afm_OLEMonitor: Tfm_OLEMonitor); virtual;
    destructor  Destroy; override;// unhook Mgr's links

    procedure   Monitoree_AddData(AName: String);
    procedure   Monitoree_Deleting;
    procedure   Monitoree_RefCountChanged(ARefCount: integer);

    function    GetGridStr(ColNum: integer; ItemNum: integer): string; virtual;
    procedure   ClearData; virtual;
    procedure   IncData(ItemNum: integer);
  end;

var
  fm_OLEMonitor: Tfm_OLEMonitor; // not always used, may use field in host form or interface object

Const
  // Standard monitor row and FData array usage:
  gwmd_Object       = 0; // name, exists
  gwmd_RefCount     = 1;
  gwmd_RefCountMax  = 2;
  gwmd_BeginObjectSpecific = 3;

//============================================================
implementation
//============================================================
uses SysUtils, Graphics;
{$R *.DFM}

Const
  Mon_ColCount = 2;
  ColWidth0 = 128;
  ColWidth1 =  96;

//--------------------------------------------------------
procedure Tfm_OLEMonitor.FormCreate(Sender: TObject);
//--------------------------------------------------------
Var
  N: integer;
begin
  // Init my data
  For N := 0 to GW_MonCount-1 do FMons[N] := nil;

  //-----------------------------------
  // init string grid how we like it
  //-----------------------------------
  sg_OLEMonitor.ColCount := Mon_ColCount;
  sg_OLEMonitor.FixedCols := 0;
  sg_OLEMonitor.RowCount := 2;
  sg_OLEMonitor.FixedRows := 1;
  sg_OLEMonitor.ColWidths[0] := ColWidth0;
  sg_OLEMonitor.ColWidths[1] := ColWidth1;
  //Fsg_OLEMonitor.ColWidths[2] := 96;
  sg_OLEMonitor.Cells[0,0] := 'Item';
  sg_OLEMonitor.Cells[1,0] := 'Value';
  // Fsg_OLEMonitor.Cells[2,0] := 'String';

  // get my title
  If Owner is TForm then  Caption := (Owner as TForm).Caption + ' Monitor'
    else Caption := 'OLE Monitor';
  DockMe;
  Height := (Owner as TForm).Height;
  Width  := ColWidth0 + ColWidth1 + 16;
end;

//--------------------------------------------------------
procedure Tfm_OLEMonitor.FormDestroy(Sender: TObject);
//--------------------------------------------------------
Var
  N: integer;
begin
  // get rid of all Mons:
  For N := 0 to GW_MonCount-1 do
  Begin
    If Assigned(FMons[N]) then FMons[N].Destroy;
  end;
end;

//--------------------------------------------------------
procedure Tfm_OLEMonitor.FormClose(Sender: TObject; var Action: TCloseAction);
//--------------------------------------------------------
begin
  Action := caHide;
end;

//--------------------------------------------------------
procedure Tfm_OLEMonitor.DockMe;
//--------------------------------------------------------
Begin
  Left := (Owner as TForm).Left + (Owner as TForm).Width;
  Top  := (Owner as TForm).Top;
end;

//--------------------------------------------------------
procedure Tfm_OLEMonitor.tm_UpdateTimer(Sender: TObject);
//--------------------------------------------------------
begin
  // Show that we're awake and getting attention
  if sh_Update.Brush.Color = clLime then
    sh_Update.Brush.Color := clBlack else
    sh_Update.Brush.Color := clLime;

  sh_Update.Update;  // force repaint

  // Update counts
  If Visible then GridUpdate;
end;

//--------------------------------------------------------
procedure Tfm_OLEMonitor.bt_ClearClick(Sender: TObject);
//--------------------------------------------------------
begin
  ClearDeadMonitors;
  ClearData;
end;

//--------------------------------------------------------
function Tfm_OLEMonitor.Mon_Add(AMon: TGW_Mon): integer; { returns monitor "handle" }
//--------------------------------------------------------
Var
  N: integer;
Label found_slot, the_end;
Begin
  Result := -1; { failed }
  For N := 0 to GW_MonCount-1 do
  Begin
    if not Assigned(FMons[N]) then goto found_slot;
  end;
  goto the_end;
found_slot:
  FMons[N] := AMon;
  Result := N;
the_end:
end;

//--------------------------------------------------------
procedure Tfm_OLEMonitor.Mon_Deleting(MonId: integer);
//--------------------------------------------------------
Begin
  FMons[MonId] := nil;
end;

//--------------------------------------------------------
procedure Tfm_OLEMonitor.ClearData;
//--------------------------------------------------------
Var
  N: integer;
Begin
  For N := 0 to GW_MonCount-1 do
  Begin
    if Assigned(FMons[N]) then FMons[N].ClearData;
  end;
end;

//--------------------------------------------------------
procedure Tfm_OLEMonitor.ClearDeadMonitors;
//--------------------------------------------------------
Var
  N: integer;
Begin
  // get rid of Mons whose Monitorees have been deleted:
  For N := 0 to GW_MonCount-1 do
  Begin
    If Assigned(FMons[N]) then
      if not Assigned(FMons[N].FMonitoree) then FMons[N].Destroy;
  end;
end;

//--------------------------------------------------------
procedure Tfm_OLEMonitor.GridUpdate;
//--------------------------------------------------------
Var
  MyRowCount: integer;
  MonNum, ItemNum, RowNum, ColNum: integer;
Begin
  // Calculate rows, and adjust grid if necessary
  MyRowCount := 1; // title row
  For MonNum := 0 to GW_MonCount-1 do
  Begin
    If Assigned(FMons[MonNum]) then MyRowCount := MyRowCount + FMons[MonNum].FDataCount;
  end;
  If MyRowCount < 2 then MyRowCount := 2; // because must be >= FixedRows+1
  If MyRowCount <> sg_OLEMonitor.RowCount then
  Begin
    sg_OLEMonitor.RowCount := MyRowCount;
    For ColNum := 0 to Mon_ColCount-1 do sg_OLEMonitor.Cells[ColNum,1] := ''; // kludge
  end;

  // Put strings in grid
  RowNum := 0;
  For MonNum := 0 to GW_MonCount-1 do
  Begin
    If Assigned(FMons[MonNum]) then
    Begin
      For ItemNum := 0 to FMons[MonNum].FDataCount -1 do
      Begin
        Inc(RowNum);
        For ColNum := 0 to Mon_ColCount-1 do
        Begin
          sg_OLEMonitor.Cells[ColNum,RowNum] := FMons[MonNum].GetGridStr(ColNum, ItemNum);
        end; { for }
      end; { for }
    end; { if }
  end; { for }
end;

//--------------------------------------------------------
procedure Tfm_OLEMonitor.DeleteCOMObject(id: integer);
//--------------------------------------------------------
Var
  MonNum: integer;
Begin
  For MonNum := 0 to GW_MonCount-1 do
  Begin
    If Assigned(FMons[MonNum]) then
    Begin
      if Assigned(FMons[MonNum].FMonitoree) then
      Begin
        if FMons[MonNum].FMonitoree.FMyObjectId = id then
          FMons[MonNum].FMonitoree.Destroy;
      end;
    end;
  end;
end;

//--------------------------------------------------------
procedure Tfm_OLEMonitor.DeleteAllCOMObjects;
//--------------------------------------------------------
Var
  MonNum: integer;
Begin
  For MonNum := 0 to GW_MonCount-1 do
  Begin
    If Assigned(FMons[MonNum]) then
    Begin
      if Assigned(FMons[MonNum].FMonitoree) then
      Begin
        FMons[MonNum].FMonitoree.Destroy;
      end;
    end;
  end;
end;

//=======================================================
//        TGW_MonMgr
//=======================================================

//---------------------------------------
//             TGW_Mon
//---------------------------------------

//--------------------------------------------------------
constructor TGW_Mon.Create(AMonitoree: TGWMonInterfacedObject; Afm_OLEMonitor: Tfm_OLEMonitor);
//--------------------------------------------------------
Var N: integer;
Begin
  inherited Create;
  Ffm_OLEMonitor := Afm_OLEMonitor;
  FMonitoree  := AMonitoree;
  FMonId      := -1;
  For N := 0 to GW_MonDataCount-1 do
  Begin
    FData[N]    := 0;
    FRowName[N] := '---';
  end;
  FDataCount    := 0;

  // standard data init
  FRowName[0] := FMonitoree.ClassName;
  FData[0]    := FMonitoree.FMyObjectId;

  FRowName[1] := 'RefCount';
  FRowName[2] := 'Max RefCount';

  FDataCount  := 3;
  // creator should add rows as needed

  // add me to mgr
  FMonId      := Ffm_OLEMonitor.Mon_Add(self);
end;

//--------------------------------------------------------
destructor  TGW_Mon.Destroy; // unhook Mgr's links
//--------------------------------------------------------
Begin
  // tell MonMgr to forget me
  If Assigned(Ffm_OLEMonitor) then Ffm_OLEMonitor.Mon_Deleting(FMonId);
  If Assigned(FMonitoree) then FMonitoree.Mon_Deleting;
  inherited Destroy;
end;

//--------------------------------------------------------
procedure TGW_Mon.Monitoree_AddData(AName: String);
//--------------------------------------------------------
Begin
  FRowName[FDataCount] := AName;
  Inc(FDataCount);
end;

//--------------------------------------------------------
procedure TGW_Mon.Monitoree_Deleting;
//--------------------------------------------------------
Begin
  FMonitoree := nil;
end;

//--------------------------------------------------------
procedure TGW_Mon.Monitoree_RefCountChanged(ARefCount: integer);
//--------------------------------------------------------
Begin
  FData[gwmd_RefCount] := ARefCount;
  If ARefCount > FData[gwmd_RefCountMax] then
    FData[gwmd_RefCountMax] := ARefCount;
end;
//--------------------------------------------------------
procedure TGW_Mon.ClearData;
//--------------------------------------------------------
Var N: integer;
Begin
  For N := gwmd_RefCountMax to GW_MonDataCount-1 do
    FData[N] := 0;
end;

//--------------------------------------------------------
procedure TGW_Mon.IncData(ItemNum: integer);
//--------------------------------------------------------
Begin
  If (ItemNum >= 0) and (ItemNum < GW_MonDataCount) then Inc(FData[ItemNum]);
end;

//--------------------------------------------------------
function TGW_Mon.GetGridStr(ColNum: integer; ItemNum: integer): string;
//--------------------------------------------------------
Var
  S: string;
Begin
  S := '---';
  Case ColNum of
    0: Begin
      S := FRowName[ItemNum];
      if ItemNum <> gwmd_Object then S := '    ' + S;
    end;
    1: Begin
      S := IntToStr(FData[ItemNum]);
      If ItemNum = gwmd_Object then
      Begin
        S := 'Id: ' + S;
        if Assigned(FMonitoree) then S := S + ' Exists'
          else S := S + ' Deleted';
      end;
    end; { 1 }
  end; { case ColNum }
  Result := S;
end;

//--------------------------------------------------------
//  TGWInterfacedObject
//--------------------------------------------------------
function TGWInterfacedObject.QueryInterface(const IID: TGUID; out Obj): Integer;
const
  E_NOINTERFACE = $80004002;
begin
  if GetInterface(IID, Obj) then Result := 0 else Result := E_NOINTERFACE;
end;

function TGWInterfacedObject._AddRef: Integer;
begin
  Inc(FRefCount);
  Result := FRefCount;
  RefCountChanged;
end;

function TGWInterfacedObject._Release: Integer;
begin
  Dec(FRefCount);
  RefCountChanged;
  if FRefCount = 0 then
  begin
    Destroy;
    Result := 0;
    Exit;
  end;
  Result := FRefCount;
end;

procedure TGWInterfacedObject.RefCountChanged;
Begin // do nothing
end;

//--------------------------------------------------------
//          TGWMonInterfacedObject
//--------------------------------------------------------

Var
  TGWMonInterfacedObject_LastObjectId: integer;

//--------------------------------------------------------
constructor TGWMonInterfacedObject.Create(Afm_OLEMonitor: Tfm_OLEMonitor); // or nil for no monitoring
//--------------------------------------------------------
Begin
  inherited Create;
  Inc(TGWMonInterfacedObject_LastObjectId);
  If TGWMonInterfacedObject_LastObjectId >= MaxInt-10 then
     TGWMonInterfacedObject_LastObjectId := 0;
  FMyObjectId := TGWMonInterfacedObject_LastObjectId;

  if Afm_OLEMonitor = nil then FMon := nil else
  Begin
    FMon := TGW_Mon.Create(self, Afm_OLEMonitor);
  end;
end;

//--------------------------------------------------------
destructor TGWMonInterfacedObject.Destroy;
//--------------------------------------------------------
Begin
  If Assigned(FMon) then FMon.Monitoree_Deleting;
  inherited Destroy;
end;
//--------------------------------------------------------
procedure TGWMonInterfacedObject.Mon_Deleting;
//--------------------------------------------------------
Begin
  FMon := nil;
end;

//--------------------------------------------------------
procedure TGWMonInterfacedObject.Mon_IncData(ItemNum: integer);
//--------------------------------------------------------
Begin
  If Assigned(FMon) then FMon.IncData(ItemNum);
end;

//--------------------------------------------------------
procedure TGWMonInterfacedObject.RefCountChanged;
//--------------------------------------------------------
Begin
  If Assigned(FMon) then FMon.Monitoree_RefCountChanged(RefCount);
end;

initialization
  TGWMonInterfacedObject_LastObjectId := 0;
end.
